/*
 * Cresques Mapping Suite. Graphic Library for constructing mapping applications.
 *
 * Copyright (C) 2004-5.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,USA.
 *
 * For more information, contact:
 *
 * cresques@gmail.com
 */
package org.cresques.ui.raster;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.image.DataBuffer;
import java.io.File;
import java.util.Vector;

import javax.swing.AbstractCellEditor;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;

import org.cresques.io.GeoRasterFile;


/**
 * Selecciona las bandas visibles en un raster. Contiene una tabla con una fila por cada
 * banda de la imagen. Por medio de checkbox se selecciona para cada RGB que banda de la 
 * imagen ser visualizada. 
 * @author Luis W. Sevilla (sevilla_lui@gva.es)
 * @author Nacho Brodin (brodin_ign@gva.es)
 */
public class BandSetupPanel extends JPanel implements TableModelListener,
                                                      KeyListener,
                                                      ActionListener,
                                                      ComponentListener,
                                                      IResize{
    final private static long serialVersionUID = -3370601314380922368L;

    /**
     * Variables para la asignacin de tamao de los componentes del panel.
     */
    private int						wComp = 445, hComp = 239;
    private int						wFileList = wComp, hFileList = (int)Math.round(hComp * 0.46);	
    private int						wBand = wFileList, hBand = hComp - hFileList - 20;
    
    /**
     * Asigna la banda del rojo
     */
    public static final int RED_BAND = GeoRasterFile.RED_BAND;

    /**
     * Asigna la banda del verde
     */
    public static final int GREEN_BAND = GeoRasterFile.GREEN_BAND;

    /**
     * Asigna banda del azul
     */
    public static final int BLUE_BAND = GeoRasterFile.BLUE_BAND;
    private final static String[] columnNames = { "R", "G", "B", "Band" };

    /**
     * Nombre del panel
     */
    private String nom = "Bandas";
    private FileList fileList = null;
    private JTable rgbTable = null;
    private JScrollPane rgbBandAssignPane = null;
    RGBBandAsignTableModel tableModel = null;
    private int sizeX = 445;
    private int sizeY = 174;
    private byte mode = 3;
    private int[] col = { 0, 1 }; //Ultima y penultima columnas activadas del jtable para cuando hay 2 bandas seleccionadas en el combo

    /**
     * This method initializes
     *
     */
    public BandSetupPanel() {
        super();
        initialize();
    }

    /**
     * This method initializes this
     *
     * @return void
     */
    void initialize() {
        //this.setPreferredSize(new Dimension(wComp, hComp));
        this.setLayout(null);
        this.setLocation(0, 0);
        this.add(getFileList(), null);
        this.add(getRGBBandAssignPane(), null);
        getFileList().getJComboBox().addKeyListener(this);
        getFileList().getJComboBox().addActionListener(this);
        this.addComponentListener(this);
        this.setComponentSize(wComp, hComp);
    }

    /**
     * Obtiene el nombre del panel
     * @return Cadena con el nombre del panel
     */
    public String getName(){
    	return this.nom;
    }
    
    /**
     * Aade la lista de georasterfiles a la tabla
     * @param files
     */
    public void addFiles(GeoRasterFile[] files) {
        for (int i = 0; i < files.length; i++) {
            String fName = files[i].getName();
            getFileList().addFName(fName);

            //((DefaultListModel)fileList.getJList().getModel()).addElement(fName);
            String bandName = new File(fName).getName();
            String bandType = "";

            switch (files[i].getDataType()) {
            case DataBuffer.TYPE_BYTE:
                bandType = "8U";
                break;
            case DataBuffer.TYPE_INT:
                bandType = "32";
                break;
            case DataBuffer.TYPE_DOUBLE:
                bandType = "64";
                break;
            case DataBuffer.TYPE_FLOAT:
                bandType = "32";
                break;
            case DataBuffer.TYPE_SHORT:
                bandType = "16";
                break;
            case DataBuffer.TYPE_USHORT:
                bandType = "16U";
                break;
            case DataBuffer.TYPE_UNDEFINED:
                bandType = "??";
                break;
            }

            if (files[i].getBandCount() > 1) {
                for (int b = 0; b < files[i].getBandCount(); b++)
                    addBand((b + 1) + " [" + bandType + "] " + bandName);
            } else {
                addBand("1 [" + bandType + "] " + bandName);
            }
        }
    }

    /**
     * Elimina un fichero de la lista
     * @param file Nombre del fichero a eliminar
     */
    public void removeFile(String file) {
        getFileList().removeFName(file);

        for (int i = 0;
                 i < ((DefaultTableModel) rgbTable.getModel()).getRowCount();
                 i++) {
            //Si el fichero borrado estaba seleccionado como banda visible. Pasaremos
            //esta visibilidad a la banda inmediata superior y si esta accin produce 
            //una excepcin (porque no hay) se pasa al inmediato inferior.
            if (((String) ((DefaultTableModel) rgbTable.getModel()).getValueAt(i, 3)).endsWith(file)) {
                try {
                    if (((Boolean) ((DefaultTableModel) rgbTable.getModel()).getValueAt(i, 0)).booleanValue()) {
                        ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), i - 1, 0);
                    }
                } catch (ArrayIndexOutOfBoundsException exc) {
                    ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), i + 1, 0);
                }

                try {
                    if (((Boolean) ((DefaultTableModel) rgbTable.getModel()).getValueAt(i, 1)).booleanValue()) {
                        ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), i - 1, 1);
                    }
                } catch (ArrayIndexOutOfBoundsException exc) {
                    ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), i + 1, 1);
                }

                try {
                    if (((Boolean) ((DefaultTableModel) rgbTable.getModel()).getValueAt(i, 2)).booleanValue()) {
                        ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), i - 1, 2);
                    }
                } catch (ArrayIndexOutOfBoundsException exc) {
                    ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), i + 1, 2);
                }

                ((DefaultTableModel) rgbTable.getModel()).removeRow(i);
                i--; //Ojo! que hemos eliminado una fila
            }
        }
    }

    /**
    * Cuando cambiamos el combo de seleccion de numero de bandas a visualizar
    * debemos resetear la tabla de checkbox para que no haya activados ms de
    * los permitidos
    * @param mode
    */
    private void resetMode(int mode) {
        //Reseteamos los checkbox
        for (int i = 0; i < (rgbTable.getColumnCount() - 1); i++)
            for (int j = 0; j < rgbTable.getRowCount(); j++)
                ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(false), j, i);

        //Asignamos los valores
        if (this.getNBands() >= 3) {
            switch (mode) {
            case 3:
                ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), 2, 2);
            case 2:
                ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), 1, 1);
            case 1:
                ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), 0, 0);
            }
        } else if (this.getNBands() == 2) {
            switch (mode) {
            case 3:
                ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), 1, 2);
            case 2:
                ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), 1, 1);
            case 1:
                ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), 0, 0);
            }
        } else if (this.getNBands() == 1) {
            switch (mode) {
            case 3:
                ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), 0, 2);
            case 2:
                ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), 0, 1);
            case 1:
                ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(true), 0, 0);
            }
        }

        col[0] = 0;
        col[1] = 1;
    }

    /**
     * Asigna modo 1, 2, o 3 bandas. El modo 1 solo permite seleccionar en la
     * tabla un checkbox, el 2 dos checkbox en distintar bandas y el 3 tres checkbox
     * tambin en distintas bandas.
     * @param mode
     */
    private void setMode(int mode) {
        //Solo hay un checkbox activado
        if (mode == 1) {
            for (int i = 0; i < rgbTable.getRowCount(); i++)
                for (int j = 0; j < (rgbTable.getColumnCount() - 1); j++) {
                    if ((i != rgbTable.getSelectedRow()) ||
                            (j != rgbTable.getSelectedColumn())) {
                        ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(false), i, j);
                    }
                }

            //Hay dos checkbox activados
        } else if (mode == 2) {
            int n = 0;

            for (int i = 0; i < (rgbTable.getColumnCount() - 1); i++)
                for (int j = 0; j < rgbTable.getRowCount(); j++)
                    if (((Boolean) ((DefaultTableModel) rgbTable.getModel()).getValueAt(j, i)).booleanValue()) {
                        n++;
                    }

            //Si se ha seleccionado 3 bandas hay eliminar una de ellas. Siempre ser la ms antigua que se clickeo
            if (n > 2) {
                for (int i = 0; i < rgbTable.getRowCount(); i++)
                    ((DefaultTableModel) rgbTable.getModel()).setValueAt(new Boolean(false), i, col[0]);
            }

            //Rotamos el punto pinchado ms antiguo para que se eliminen alternativamente
            if ((col[0] == rgbTable.getSelectedColumn()) ||
                    ((col[0] != rgbTable.getSelectedColumn()) &&
                    (col[1] != rgbTable.getSelectedColumn()))) {
                col[0] = col[1];
                col[1] = rgbTable.getSelectedColumn();
            }

            //El modo 3 es el comportamiento original
        } else if (mode == 3) {
            return;
        }
    }

    /**
     * Obtiene el panel que contiene la lista de ficheros por banda.
     * @return Panel FileList
     */
    public FileList getFileList() {
        if (fileList == null) {
            fileList = new FileList();
            //fileList.setBounds(9, 9, wFileList, hFileList);
        }

        return fileList;
    }

    /**
     * This method initializes jTable
     *
     * @return javax.swing.JTable
     */
    private JScrollPane getRGBBandAssignPane() {
        if (rgbBandAssignPane == null) {
            rgbBandAssignPane = new JScrollPane(getRGBTable());
            //rgbBandAssignPane.setBounds(10, 123, wBand, hBand);

            TableColumn column = null;

            for (int i = 0; i < 3; i++) {
                column = rgbTable.getColumnModel().getColumn(i);
                column.setCellRenderer(new RadioColumnRenderer());
                column.setCellEditor(new RadioColumnEditor());
                column.setMaxWidth(22);
                column.setMinWidth(22);
            }

            //			frame.setContentPane(scrollPane);
        }

        return rgbBandAssignPane;
    }

    /**
     * Obtiene la Tabla
     * @return Tabla de bandas de la imagen
     */
    public JTable getRGBTable() {
        if (rgbTable == null) {
            tableModel = new RGBBandAsignTableModel();
            tableModel.addTableModelListener(this);
            rgbTable = new JTable(tableModel);
            rgbTable.setPreferredScrollableViewportSize(new Dimension(328, 72));
        }

        return rgbTable;
    }

    /**
     * Asigna al combo de nmero de bandas a mostrar el valor correspondiente
     * dependiendo del nmero de bandas de la imagen.
     */
    public void setList() {
        String[] l = null;

        if (this.getNBands() == 1) {
            l = new String[1];
            l[0] = "1";
        }

        if (this.getNBands() == 2) {
            l = new String[2];
            l[0] = "1";
            l[1] = "2";
        }

        if (this.getNBands() == 3) {
            l = new String[3];
            l[0] = "1";
            l[1] = "2";
            l[2] = "3";
        }

        if (this.getNBands() > 0) {
            getFileList().setList(l);
        }
    }

    /**
     * Aade una banda a la tabla  bandas de la imagen asignandole un
     * nombre y valor a los checkbox
     * @param bandName        Nombre de la banda
     */
    private void addBand(String bandName) {
        Vector v = new Vector();
        v.add(new Boolean(false));
        v.add(new Boolean(false));
        v.add(new Boolean(false));
        v.add(bandName);
        ((DefaultTableModel) rgbTable.getModel()).addRow(v);
    }

    /**
     * Obtiene el nmero de bandas de la lista
     * @return
     */
    public int getNBands() {
        return ((DefaultTableModel) rgbTable.getModel()).getRowCount();
    }

    /**
     * Obtiene el nombre de la banda de la posicin i de la tabla
     * @param i
     * @return
     */
    public String getBandName(int i) {
        String s = (String) ((DefaultTableModel) rgbTable.getModel()).getValueAt(i, 3);
        return s.substring(s.lastIndexOf("[8U]") + 5, s.length());
    }

    /**
     * Mantiene la asignacin entre R, G o B y la banda de la imagen que le corresponde 
     * @param nBand	Banda de la imagen que corresponde
     * @param flag	R, G o B se selecciona por medio de un flag que los identifica
     */
    public void assignBand(int nBand, int flag) {
        Boolean t = new Boolean(true);

        if ((flag & RED_BAND) == RED_BAND)
            ((DefaultTableModel) rgbTable.getModel()).setValueAt(t, nBand, 0);
        
        if ((flag & GREEN_BAND) == GREEN_BAND) 
            ((DefaultTableModel) rgbTable.getModel()).setValueAt(t, nBand, 1);
        
        if ((flag & BLUE_BAND) == BLUE_BAND)
            ((DefaultTableModel) rgbTable.getModel()).setValueAt(t, nBand, 2);
    }

    /**
     * Obtiene la correspondencia entre el R, G o B y la banda asignada
     * @param flag	R, G o B se selecciona por medio de un flag que los identifica
     * @return	Banda de la imagen asignada al flag pasado por parmetro
     */
    public int getAssignedBand(int flag) {
        if ((flag & RED_BAND) == RED_BAND) {
            for (int nBand = 0; nBand < rgbTable.getRowCount(); nBand++)
                if (((Boolean) ((DefaultTableModel) rgbTable.getModel()).getValueAt(nBand,
                                                                                        0)).booleanValue()) {
                    return nBand;
                }
        }

        if ((flag & GREEN_BAND) == GREEN_BAND) {
            for (int nBand = 0; nBand < rgbTable.getRowCount(); nBand++)
                if (((Boolean) ((DefaultTableModel) rgbTable.getModel()).getValueAt(nBand,
                                                                                        1)).booleanValue()) {
                    return nBand;
                }
        }

        if ((flag & BLUE_BAND) == BLUE_BAND) {
            for (int nBand = 0; nBand < rgbTable.getRowCount(); nBand++)
                if (((Boolean) ((DefaultTableModel) rgbTable.getModel()).getValueAt(nBand,
                                                                                        2)).booleanValue()) {
                    return nBand;
                }
        }

        return -1;
    }

    /* (non-Javadoc)
     * @see javax.swing.event.TableModelListener#tableChanged(javax.swing.event.TableModelEvent)
     */
    public void tableChanged(TableModelEvent e) {
        rgbTable.revalidate();
        rgbBandAssignPane.revalidate();
        revalidate();
    }

    /* (non-Javadoc)
     * @see java.awt.event.KeyListener#keyPressed(java.awt.event.KeyEvent)
     */
    public void keyPressed(KeyEvent arg0) {
        // TODO Auto-generated method stub
    }

    /* (non-Javadoc)
     * @see java.awt.event.KeyListener#keyReleased(java.awt.event.KeyEvent)
     */
    public void keyReleased(KeyEvent arg0) {
        // TODO Auto-generated method stub
    }

    /* (non-Javadoc)
     * @see java.awt.event.KeyListener#keyTyped(java.awt.event.KeyEvent)
     */
    public void keyTyped(KeyEvent arg0) {
        // TODO Auto-generated method stub
    }

    /**
     *Evento que activado cuando se modifica el nmero de bandas que se desean visualizar.
     *Se obtiene este nmero y se llama a la funcin que controla el cambio de nmero
     *de bandas visualizadas. 
     */
    public void actionPerformed(ActionEvent arg0) {
        String vBands = (String) getFileList().getJComboBox().getSelectedItem();

        if (vBands != null) {
            if (vBands.compareTo("3") == 0) {
                mode = 3;
            }

            if (vBands.compareTo("2") == 0) {
                mode = 2;
            }

            if (vBands.compareTo("1") == 0) {
                mode = 1;
            }

            resetMode(mode);
        }
    }

    class RadioColumnEditor extends AbstractCellEditor
        implements TableCellEditor {
        final private static long serialVersionUID = -3370601314380922368L;
        public JRadioButton theRadioButton;

        public RadioColumnEditor() {
            super();
            theRadioButton = new JRadioButton();
            theRadioButton.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent event) {
                        fireEditingStopped() ;
                        setMode(mode);
                    }
                });
        }

        public Component getTableCellEditorComponent(JTable table, Object obj,
                                                     boolean isSelected,
                                                     int row, int col) {
            theRadioButton.setHorizontalAlignment(SwingUtilities.CENTER);

            Boolean lValueAsBoolean = (Boolean) obj;
            theRadioButton.setSelected(lValueAsBoolean.booleanValue());

            return theRadioButton;
        }

        public Object getCellEditorValue() {
            return new Boolean(theRadioButton.isSelected());
        }
    }

    class RadioColumnRenderer extends JRadioButton implements TableCellRenderer {
        final private static long serialVersionUID = -3370601314380922368L;

        public Component getTableCellRendererComponent(JTable table,
                                                       Object value,
                                                       boolean isSelected,
                                                       boolean hasFocus,
                                                       int row, int column) {
            if (value == null) {
                this.setSelected(false);
            }

            Boolean ValueAsBoolean = (Boolean) value;
            this.setSelected(ValueAsBoolean.booleanValue());
            this.setHorizontalAlignment(SwingConstants.CENTER);

            return this;
        }
    }

    class RGBBandAsignTableModel extends DefaultTableModel {
        final private static long serialVersionUID = -3370601314380922368L;

        public RGBBandAsignTableModel() {
            super(new Object[0][4], columnNames);
        }

        public Class getColumnClass(int c) {
            if (c < 3) {
                return Boolean.class;
            }

            return String.class;
        }

        public void setValueAt(Object value, int row, int col) {
            if ((col < 3) && ((Boolean) value).booleanValue()) {
                for (int i = 0; i < getRowCount(); i++) {
                    if (i != row) {
                        setValueAt(new Boolean(false), i, col);
                    }
                }
            }

            super.setValueAt(value, row, col);
        }

        public void addNew() {
            super.addRow(new Object[] {
                             new Boolean(false), new Boolean(false),
                             new Boolean(false), ""
                         });
        }
        
              
    }

    
    public void setComponentSize(int w, int h){
    	wComp = w; hComp = h;
        wFileList = wComp - 18; hFileList = (int)Math.round(hComp * 0.46);	
        wBand = wFileList; hBand = hComp - hFileList - 20;
        
        this.setPreferredSize(new Dimension(wComp, hComp));
        this.setSize(wComp, hComp);
        rgbBandAssignPane.setBounds(10, hFileList + 12, wBand, hBand);
        fileList.setBounds(9, 9, wFileList, hFileList);
        
    }
    
    
    
	public void componentHidden(ComponentEvent e) {
		// TODO Auto-generated method stub
		
	}

	public void componentMoved(ComponentEvent e) {
		// TODO Auto-generated method stub
		
	}
	/**
	 * Redimensiona el panel cuando se redimensiona el contenedor de ste
	 */
	public void componentResized(ComponentEvent e) {
		/*if(e.getSource() == this){
			int nWidth = this.getSize().width;
			int nHeight = this.getSize().height;
			int difWidth = nWidth - 445;
			int difHeight = nHeight - 239;
			this.fileList.setResize(difWidth, difHeight);
			this.rgbBandAssignPane.setBounds(10, 123 + difHeight/2, 427 + difWidth, 104 + difHeight/2);
			
		}*/
		
	}

	public void componentShown(ComponentEvent e) {
		// TODO Auto-generated method stub
		
	}
} //  @jve:decl-index=0:visual-constraint="10,10"
