package com.hardcode.gdbms.engine.data.file;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import com.hardcode.gdbms.engine.data.db.FakeTransactionSupport;
import com.hardcode.gdbms.engine.data.driver.DriverException;
import com.hardcode.gdbms.engine.data.driver.FileDriver;
import com.hardcode.gdbms.engine.data.edition.EditionInfo;
import com.hardcode.gdbms.engine.data.edition.PKTable;
import com.hardcode.gdbms.engine.values.Value;
import com.hardcode.gdbms.engine.values.ValueCollection;
import com.hardcode.gdbms.engine.values.ValueFactory;

/**
 * @author Fernando Gonzlez Corts
 */
public class FileDataWareImpl extends FileDataSourceAdapter implements
        FileDataWare {

    private FakeTransactionSupport ftSupport = new FakeTransactionSupport(this);

    public void start() throws DriverException {
        throw new RuntimeException("Invoke beginTrans in a DataWare");
    }

    public void stop() throws DriverException {
        throw new RuntimeException(
                "Invoke commitTrans/rollBackTrans in a DataWare");
    }

    public void beginTrans() throws DriverException {
        super.start();
        ftSupport.beginTrans();
    }

    public void commitTrans() throws DriverException {
        ((FileDriver)getDriver()).writeFile(this);
        super.stop();
        ftSupport.commitTrans();
    }

    public void copy(File in, File out) throws IOException {
        FileInputStream fis = new FileInputStream(in);
        FileOutputStream fos = new FileOutputStream(out);
        byte[] buf = new byte[1024];
        int i = 0;
        while ((i = fis.read(buf)) != -1) {
            fos.write(buf, 0, i);
        }
        fis.close();
        fos.close();
    }

    public void deleteRow(long rowId) throws DriverException {
        ftSupport.deleteRow(rowId);
    }

    public void insertEmptyRow(ValueCollection pk) throws DriverException {
        ftSupport.insertEmptyRow(pk);
    }

    public void insertFilledRow(Value[] values) throws DriverException {
        ftSupport.insertFilledRow(values);
    }

    public void rollBackTrans() throws DriverException {
        super.stop();
        ftSupport.rollBackTrans();
    }

    public void setFieldValue(long row, int fieldId, Value value)
            throws DriverException {
        ValueCollection pkValue = null;
        Value[] originalRow = null;
        if (row < super.getRowCount()) {
            pkValue = getOriginalPKValue(row);
            originalRow = getOriginalRow(row);
        }
        ftSupport.setFieldValue(row, fieldId, value, pkValue, originalRow);
    }

    /**
     * @see com.hardcode.gdbms.engine.data.driver.ReadAccess#getFieldValue(long,
     *      int)
     */
    public Value getFieldValue(long rowIndex, int fieldId)
            throws DriverException {
        // If there is no transaction
        if (!ftSupport.isTransaction()) {
            return super.getFieldValue(rowIndex, fieldId);
        }

        // get where's the row
        EditionInfo fip = ftSupport.getPKTable().getIndexLocation(
                (int) rowIndex);
        int flag = fip.getFlag();

        // get the value
        if ((flag == PKTable.ADDED) || (flag == PKTable.MODIFIED)) {
            return ftSupport.getInternalBuffer().getFieldValue(fip.getIndex(),
                    fieldId);
        } else {
            return super.getFieldValue(fip.getIndex(), fieldId);
        }
    }

    /**
     * @see com.hardcode.gdbms.engine.data.driver.ReadAccess#getRowCount()
     */
    public long getRowCount() throws DriverException {
        if (ftSupport.isTransaction()) {
            // If there is a transaction
            return ftSupport.getPKTable().getPKCount();
        } else {
            // If there's not
            return super.getRowCount();
        }
    }

    /**
     * Gets the value of the original row
     * 
     * @param rowIndex
     *            index of the row to be retrieved
     * 
     * @return Row values
     * 
     * @throws DriverException
     *             if the operation fails
     */
    private Value[] getOriginalRow(long rowIndex) throws DriverException {
        Value[] ret = new Value[getFieldCount()];

        for (int i = 0; i < ret.length; i++) {
            ret[i] = super.getFieldValue(rowIndex, i);
        }

        return ret;
    }

    /**
     * Gets the primary key value in the original DataSource
     * 
     * @param rowIndex
     *            index of the row to be read
     * 
     * @return PK
     * 
     * @throws DriverException
     *             If the operation fails
     */
    private ValueCollection getOriginalPKValue(long rowIndex)
            throws DriverException {
        int[] fieldsId = getPrimaryKeys();
        Value[] pks = new Value[fieldsId.length];

        for (int i = 0; i < pks.length; i++) {
            pks[i] = super.getFieldValue(rowIndex, fieldsId[i]);
        }

        return ValueFactory.createValue(pks);
    }
}
