package com.hardcode.gdbms.engine.data.indexes;

import java.io.IOException;


/**
 * Implementacin del conjunto de ndices que guarda en memoria hasta un lmite
 * y  a partir de ese lmite pasa todos los ndices a disco
 *
 * @author Fernando Gonzlez Corts
 */
public class IndexSetImpl implements VariableIndexSet {
	private VariableIndexSet set;
	private boolean complete = false;

	/**
	 * Nmero de ndices que se guardan en memoria, a partir de dicho nmero se
	 * guarda en disco
	 */
	private int limit;

	/**
	 * Creates a new IndexSetImpl object.
	 */
	public IndexSetImpl() {
		limit = IndexFactory.MEMORY_THRESHOLD;
		set = new VariableMemoryIndexSet(IndexFactory.MEMORY_THRESHOLD);
	}

	/**
	 * Creates a new IndexSetImpl object.
	 *
	 * @param limit lmite a partir del cual se guardan todos los ndices en
	 * 		  disco
	 */
	public IndexSetImpl(int limit) {
		this.limit = limit;
		set = new VariableMemoryIndexSet(limit);
	}

	/**
	 * Aade un ndice al conjunto
	 *
	 * @param index ndice a aadir
	 *
	 * @throws IOException Si se produce un error al escribir en el disco
	 * @throws RuntimeException
	 */
	public synchronized void addIndex(long index) throws IOException {
		if (complete) {
			throw new RuntimeException(
				"Cannot add more indexes after indexSetComplete");
		}

		if (set.getIndexCount() == limit) {
			//Se sustituye el ndice de memoria por el ndice de disco
			VariableDiskIndexSet newSet = new VariableDiskIndexSet();
			newSet.open();
			newSet.addAll(set);
			set = newSet;
		}

		set.addIndex(index);
	}

	/**
	 * Devuelve el ndice nth-simo si se invoc previamente a indexSetComplete
	 * y lanza una excepcin en caso contrario
	 *
	 * @param nth ndice de ndice que se quiere obtener
	 *
	 * @return indice nth-simo
	 *
	 * @throws IOException Si se produce un error accediendo a disco
	 * @throws RuntimeException
	 */
	public long getIndex(long nth) throws IOException {
		if (!complete) {
			throw new RuntimeException("Must call indexSetComplete First");
		}

		return set.getIndex(nth);
	}

	/**
	 * Devuelve el nmero de ndices si se invoc previamente a
	 * indexSetComplete y lanza una excepcin en caso contrario
	 *
	 * @return nmero de ndices
	 *
	 * @throws RuntimeException
	 */
	public long getIndexCount() {
		if (!complete) {
			throw new RuntimeException("Must call indexSetComplete First");
		}

		return set.getIndexCount();
	}

	/**
	 * @see com.hardcode.gdbms.engine.data.indexes.VariableIndexSet#indexSetComplete()
	 */
	public void indexSetComplete() throws IOException {
		complete = true;
		set.indexSetComplete();
	}

	/**
	 * @see com.hardcode.gdbms.engine.data.indexes.VariableIndexSet#open(java.io.File)
	 */
	public void open() throws IOException {
		set.open();
	}

	/**
	 * @see com.hardcode.gdbms.engine.data.indexes.VariableIndexSet#close()
	 */
	public void close() throws IOException {
		set.close();
	}

	/**
	 * DOCUMENT ME!
	 *
	 * @return DOCUMENT ME!
	 *
	 * @throws IOException
	 *
	 * @see com.hardcode.gdbms.engine.data.indexes.VariableIndexSet#getIndexes()
	 */
	public long[] getIndexes() throws IOException {
		return set.getIndexes();
	}
}
