package org.gvsig.remoteClient.wms;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.net.URL;
import java.net.URLConnection;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.TreeMap;
import java.util.Vector;

import org.gvsig.remoteClient.exceptions.ServerErrorException;
import org.gvsig.remoteClient.exceptions.WMSException;
import org.gvsig.remoteClient.utils.CapabilitiesTags;
import org.gvsig.remoteClient.utils.ExceptionTags;
import org.gvsig.remoteClient.utils.Utilities;
import org.kxml2.io.KXmlParser;
import org.xmlpull.v1.XmlPullParserException;

/**
 * <p> Abstract class that represents handlers to comunicate via WMS protocol.
 * </p>
 * 
 */
public abstract class WMSProtocolHandler {
	/**
	 * Encoding used to parse different xml documents.
	 */
	protected String encoding = "UTF-8";
	/**
	 * procotol handler name
	 */
    protected String name;
    /**
     * protocol handler version
     */
    protected String version;
    /**
     * host of the WMS to connect
     */
    protected String host;
    /**
     * port number of the comunication channel of the WMS to connect
     */
    protected String port;    
    /**
     * WMS metadata
     */
    protected ServiceInformation serviceInfo;
    public TreeMap layers;
    public WMSLayer rootLayer;
//    public Vector srs;
    
    /**
     * parses the data retrieved by the WMS in XML format. 
     * It will be mostly the WMS Capabilities, but the implementation
     * will be placed in the handler implementing certain version of the protocol.
     * 
     */
    public abstract void parse(File f) ;
    
    /**
     * returns the alfanumeric information of the layers at the specified point.
     * the diference between the other getfeatureInfo method is that this will
     * be implemented by each specific version because the XML from the server will be
     * parsed and presented by a well known structure.
     */    

    public String getName() {        
    	return name;
    } 

    public String getVersion() {        
    	return version;
    }    
    
    public ServiceInformation getServiceInformation() {        
        return serviceInfo;
    }  
    public String getHost ()
    {
    	return host;
    }
    public void setHost(String _host)
    {
    	host = _host;
    }
    public String getPort()
    {
    	return port;
    }
    public void setPort(String _port)
    {
    	port = _port;
    }
    

    /**
	 * <p>Builds a GetCapabilities request that is sent to the WMS
	 * the response will be parse to extract the data needed by the
	 * WMS client</p>
	 * @param override, if true the previous downloaded data will be overridden
	 */
    public void getCapabilities(WMSStatus status, boolean override, ICancellable cancel)
    {		
    	URL request = null;
		try
		{
			request = new URL(buildCapabilitiesRequest(status));
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		try
		{
			if (override) 
				Utilities.removeURL(request);
			File f = Utilities.downloadFile(request,"wms_capabilities.xml", cancel);
			if (f == null)
				return;
			clear();
			parse(f);
	    } catch(Exception e)
		{
			//TODO
			e.printStackTrace();
		}
    }

    private void clear() {
		layers.clear();
		serviceInfo.clear();
	}

	/**
     * <p>It will send a GetFeatureInfo request to the WMS
     * Parsing the response and redirecting the info to the WMS client</p>
     */
    public String getFeatureInfo(WMSStatus status, int x, int y, int featureCount)
    {
    	URL request = null;
    	StringBuffer output = new StringBuffer();
    	String outputFormat = new String();
    	String ServiceException = "ServiceExceptionReport";    	    	
    	StringBuffer sb = new StringBuffer();
    	sb.append("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>");
		try
		{
			//TODO:
			//pass this buildXXXRequest to the WMSProtocolHandlerXXX: The request can depend on the WMS version.
			request = new URL(buildGetFeatureInfoRequest(status, x, y));
			
			//TODO:
			//see which output format is being requested. 					
		}
		catch(Exception e)
		{
			e.printStackTrace();
			return "";
		}
		
    	try
    	{	    	    	
    		System.out.println(request.toString());	  
	    	byte[] buffer = new byte[1024*256];	    	
	    	DataInputStream is = new DataInputStream(request.openStream());	    	
	    	outputFormat = request.openConnection().getContentType();

	    	for (int i = is.read(buffer); i>0; i = is.read(buffer))
	    	{
	    		String str = new String(buffer,0,i);
	    		output.append(str);	    	
	    	}
	    		    
	    	is.close();
		    if ( (outputFormat == null) || (outputFormat.indexOf("xml") != -1)
		    		||output.toString().toLowerCase().startsWith("<?xml")
		    		||(outputFormat.indexOf("gml") != -1))
		    {
		    	int tag;
		    	KXmlParser kxmlParser = null;
		    	kxmlParser = new KXmlParser();	    
		    	kxmlParser.setInput(new StringReader(output.toString()));
		    	
		    	tag = kxmlParser.nextTag();   	
		    	if (kxmlParser.getName().compareTo(ServiceException)==0)
				{
		    		sb.append("<INFO>").append(parseException( output.toString().getBytes())).append("</INFO>");
		    		return sb.toString();									
				}
				else if (kxmlParser.getName().compareToIgnoreCase("ERROR")==0)
				{
					return output.toString();
				}
				else		
				{
					return output.toString();		
				}
					    		
// 				 while(tag != KXmlParser.END_DOCUMENT)
//				 {
//					 switch(tag)
//					 {
//						case KXmlParser.START_TAG:		
//							if (kxmlParser.getName().compareTo(ServiceException)==0)
//							{
//					    		sb.append("<INFO>").append(parseException( output.toString().getBytes())).append("</INFO>");
//					    		return sb.toString();									
//							}
//							else if (kxmlParser.getName().compareToIgnoreCase("ERROR")==0)
//								return output.toString();
//							else								
//								sb.append("<" + kxmlParser.getName() + ">\n");							
//							break;
//						case KXmlParser.END_TAG:	
//							sb.append("</" + kxmlParser.getName() + ">\n");
//							break;
//						case KXmlParser.TEXT:
//							sb.append(kxmlParser.getText());						
//						break;
//					 }
//    				 tag = kxmlParser.next();
//    			 }		    		
	    		//return sb.toString();
	    	}
	    	else
	    	{
//	    		sb.append("<INFO>").append("Info format not supported").append("</INFO>");
//	    		return sb.toString();
	    		//Para que funcione con el GetFeatureInfo Viewer generico hay que devolver:
	    		 return output.toString();
	    	}
		}
    	catch(XmlPullParserException parserEx)
    	{
    		if (output.toString().toLowerCase().indexOf("xml") != -1)
    		{
    			return output.toString().trim();
    		}
    		else
    		{
    	   		sb.append("<INFO>").append("Info format not supported").append("</INFO>");
        		return sb.toString();
    		}
    	}
    	catch(Exception e)
    	{
    		e.printStackTrace();
    		sb.append("<INFO>").append("Info format not supported").append("</INFO>");
    		return sb.toString();

    	}
    }
    /**
     * <p>Builds a GetMap request that is sent to the WMS
     * the response (image) will be redirect to the
     * WMS client</p>
     */   
    public byte[] _getMap(WMSStatus status) throws ServerErrorException, WMSException
    {        
    	URL request = null;
		try
		{
			//TODO:
			//pass this buildXXXRequest to the WMSProtocolHandlerXXX: The request can depend on the WMS version.
			request = new URL(buildMapRequest(status));
			URLConnection conn = request.openConnection();
			System.out.println(request.toString());
            String type = conn.getContentType();
            
                        	
	    	byte[] imageBytes = null;
	    	byte[] buffer = new byte[1024*256];
            InputStream is = conn.getInputStream();
	    	int readed = 0;
	    	
	    	for (int i = is.read(buffer); i>0; i = is.read(buffer)){
                // Creates a new buffer to contain the previous readed bytes and the next bunch of bytes
	    		byte[] buffered = new byte[readed+i];
	    		for (int j = 0; j < buffered.length; j++) {
	    			if (j<readed){
                        // puts the previously downloaded bytes into the image buffer
	    				buffered[j] = imageBytes[j];
	    			}
	    			else {
                        // appends the recently downloaded bytes to the image buffer.
	    				buffered[j] = buffer[j-readed];
	    			}
				}
	    		imageBytes = (byte[]) buffered.clone();
	    		readed += i;	    		
	    	}
	    	
	    	if ((type !=null && !type.subSequence(0,5).equals("image")) 
	    		||(Utilities.isTextData(imageBytes)))
	    	{	    		
               	WMSException wmsEx = null;
               	
            	String exceptionMessage = parseException(imageBytes);
                if (exceptionMessage==null)
                {
                 	String error = new String(imageBytes);
                	int pos = error.indexOf("<?xml");
                	if (pos!= -1)
                	{
                		String xml = error.substring(pos,error.length());
                		exceptionMessage = parseException(xml.getBytes());
                        if (exceptionMessage == null)
                        	exceptionMessage = new String(imageBytes);
                	}
                }
             	wmsEx = new WMSException(exceptionMessage);
            	wmsEx.setWMSMessage(new String(imageBytes));
                throw wmsEx;
            }
			return imageBytes;	    	
		}
		catch(IOException e)
		{
			e.printStackTrace();
            throw new ServerErrorException();
		}
    } 
    
    public File getLegendGraphic(WMSStatus status, String layerName, ICancellable cancel) throws ServerErrorException, WMSException
    {
    	URL request = null;
		try
		{
			request = new URL(buildGetLegendGraphicRequest(status, layerName));     
			System.out.println(request);
            File f = Utilities.downloadFile(request, "wmsGetLegendGraphic", cancel);            		    	
	    	if (f== null)
	    		return null;
            if (Utilities.isTextFile(f)) {
	    		FileInputStream fis = new FileInputStream(f);
	    		FileChannel fc = fis.getChannel();
	    		byte[] data = new byte[(int)fc.size()];
	    		ByteBuffer bb = ByteBuffer.wrap(data);
	    		fc.read(bb);
	    			    		
	    		WMSException wmsEx = null;
               	
            	String exceptionMessage = parseException(data);
                if (exceptionMessage==null)
                {
                 	String error = new String(data);
                	int pos = error.indexOf("<?xml");
                	if (pos!= -1)
                	{
                		String xml = error.substring(pos,error.length());
                		exceptionMessage = parseException(xml.getBytes());
                	}               
                    if (exceptionMessage == null)
                    	exceptionMessage = new String(data);
                	
                }
             	wmsEx = new WMSException(exceptionMessage);
            	wmsEx.setWMSMessage(new String(data));
            	Utilities.removeURL(request);
                throw wmsEx;
            }
			return f;	    	
		}
		catch(IOException e)
		{
			e.printStackTrace();
            throw new ServerErrorException();
		}
    } 
    
    public File getMap(WMSStatus status, ICancellable cancel) throws ServerErrorException, WMSException
    {        
    	URL request = null;
		try
		{
			//TODO:
			//pass this buildXXXRequest to the WMSProtocolHandlerXXX: The request can depend on the WMS version.
			request = new URL(buildMapRequest(status));
            
            File f = Utilities.downloadFile(request, "wmsGetMap", cancel);            		    	
	    	if (f== null)
	    		return null;
            if (Utilities.isTextFile(f)) {
	    		FileInputStream fis = new FileInputStream(f);
	    		FileChannel fc = fis.getChannel();
	    		byte[] data = new byte[(int)fc.size()];   // fc.size returns the size of the file which backs the channel
	    		ByteBuffer bb = ByteBuffer.wrap(data);
	    		fc.read(bb);
	    			    		
	    		WMSException wmsEx = null;
               	
            	String exceptionMessage = parseException(data);
                if (exceptionMessage==null)
                {
                 	String error = new String(data);
                	int pos = error.indexOf("<?xml");
                	if (pos!= -1)
                	{
                		String xml = error.substring(pos,error.length());
                		exceptionMessage = parseException(xml.getBytes());
//                        if (exceptionMessage == null)
//                        	exceptionMessage = new String(data);
                	}               
                    if (exceptionMessage == null)
                    	exceptionMessage = new String(data);
                	
                }
             	wmsEx = new WMSException(exceptionMessage);
            	wmsEx.setWMSMessage(new String(data));
            	
            	// Since it is an error file, It must be deleted from the cache
            	Utilities.removeURL(request);
                throw wmsEx;
            }
			return f;	    	
		}
		catch(IOException e)
		{
			e.printStackTrace();
            throw new ServerErrorException();
		}
    } 
    
    
    /* (non-Javadoc)
     * @see org.gvsig.remoteClient.wms.WMSProtocolHandler#parseException(byte[])
     */
    protected String parseException(byte[] data) {
        ArrayList errors = new ArrayList();
        KXmlParser kxmlParser = new KXmlParser();
        try
        {
            kxmlParser.setInput(new ByteArrayInputStream(data), encoding);        
            kxmlParser.nextTag();
            int tag;
            if ( kxmlParser.getEventType() != KXmlParser.END_DOCUMENT ) 
            { 
                kxmlParser.require(KXmlParser.START_TAG, null, ExceptionTags.EXCEPTION_ROOT);             
                tag = kxmlParser.nextTag();
                 while(tag != KXmlParser.END_DOCUMENT)
                 {
                     switch(tag)
                     {
                        case KXmlParser.START_TAG:
                            if (kxmlParser.getName().compareTo(ExceptionTags.SERVICE_EXCEPTION)==0){
                                String errorCode = kxmlParser.getAttributeValue("", ExceptionTags.CODE);
                                errorCode = (errorCode != null) ? "["+errorCode+"] " : "";
                                String errorMessage = kxmlParser.nextText();
                                errors.add(errorCode+errorMessage);
                            }
                            break;
                        case KXmlParser.END_TAG:                            
                            break;
                        
                     }
                     tag = kxmlParser.nextTag();
                 }
                 //kxmlParser.require(KXmlParser.END_DOCUMENT, null, null);
            }
        }
        catch(XmlPullParserException parser_ex){ 
            parser_ex.printStackTrace();
        }
        catch (IOException ioe) {           
            ioe.printStackTrace();            
        }
        String message = errors.size()>0? "" : null;
        for (int i = 0; i < errors.size(); i++) {
            message += (String) errors.get(i)+"\n";
        }
        return message;
    }
    /**
     * Builds the GetCapabilitiesRequest according to the OGC WMS Specifications
     * without a VERSION, to get the highest version than a WMS supports.
     */
    public static String buildCapabilitiesSuitableVersionRequest(String _host, String _version)
    {
		String req = new String();		
        String symbol = getSymbol(_host);
        req = req + _host + symbol + "REQUEST=GetCapabilities&SERVICE=WMS&";		
        if((_version != null) && (_version.length()>0 ))
        {
        	req += ("&VERSION=" + _version);
        }
		req += ("&EXCEPTIONS=XML");
		return req;   	
    }
    
    /**
     * Builds the GetCapabilitiesRequest according to the OGC WMS Specifications
     * @param WMSStatus
     */
    private String buildCapabilitiesRequest(WMSStatus status)
    {
		StringBuffer req = new StringBuffer();
		String symbol = null;
		
		String onlineResource;
		if (status == null || status.getOnlineResource() == null)
			onlineResource = getHost();
		else 
			onlineResource = status.getOnlineResource();
		symbol = getSymbol(onlineResource);
		
		req.append(onlineResource).append(symbol).append("REQUEST=GetCapabilities&SERVICE=WMS&");
		req.append("VERSION=").append(getVersion()).append("&EXCEPTIONS=XML");
		return req.toString();
    }
    
    /**
     * Builds the GetFeatureInfoRequest according to the OGC WMS Specifications
     */
    protected String buildGetFeatureInfoRequest(WMSStatus status, int x, int y)
    {
//    	TODO: pass by parameter the info output format?
		StringBuffer req = new StringBuffer();
		String symbol = null;

		String onlineResource;
		if (status.getOnlineResource() == null)
			onlineResource = getHost();
		else 
			onlineResource = status.getOnlineResource();
		symbol = getSymbol(onlineResource);
        
		req.append(onlineResource).append(symbol).append("REQUEST=GetFeatureInfo&SERVICE=WMS&");
		req.append("QUERY_LAYERS=").append(Utilities.Vector2CS(status.getLayerNames())); 
		req.append("&VERSION=").append(getVersion()).append("&INFO_FORMAT=application/vnd.ogc.gml&");
		req.append(getPartialQuery(status)).append("&x="+x + "&y="+y);
		//this parameter sets the max number of features per layer to be returned.
		//we set it to avoid the bug in mapserver that takes this number like max number of total features.
		req.append("&FEATURE_COUNT=10000");
       if (status.getExceptionFormat() != null) {
            req.append("&EXCEPTIONS=" + status.getExceptionFormat());
        } else {
            req.append("&EXCEPTIONS=XML");
        }
		return req.toString().replaceAll(" ", "%20");
    }    

    /**
     * Builds the GetMapRequest according to the OGC WMS Specifications
     */
    private String buildGetLegendGraphicRequest(WMSStatus status, String layerName)
    { 
    	//TODO: pass by parameter the legend output format?
		StringBuffer req = new StringBuffer();
		String symbol = null;
		String onlineResource = null;
		
		if (status.getOnlineResource() == null)
			onlineResource = getHost();
		else 
			onlineResource = status.getOnlineResource();
		symbol = getSymbol(onlineResource);
        
		req.append(onlineResource + symbol + "REQUEST=GetLegendGraphic&SERVICE=WMS&VERSION=").append(getVersion()).append("&");
        req.append("LAYER=" + layerName)
        .append("&FORMAT=image/png");
		return req.toString().replaceAll(" ", "%20");
    }

    /**
     * Builds the GetMapRequest according to the OGC WMS Specifications
     */
    private String buildMapRequest(WMSStatus status)
    { 
		StringBuffer req = new StringBuffer();
		String symbol = null;
		String onlineResource = null;
		
		if (status.getOnlineResource() == null)
			onlineResource = getHost();
		else 
			onlineResource = status.getOnlineResource();
		symbol = getSymbol(onlineResource);
        
		req.append(onlineResource + symbol + "REQUEST=GetMap&SERVICE=WMS&VERSION=").append(getVersion()).append("&");
		req.append(getPartialQuery(status));
//        if (status.getExceptionFormat() != null) {
//            req.append("&EXCEPTIONS=" + status.getExceptionFormat());
//        } else {
//            req.append("&EXCEPTIONS=XML");
//        }
		return req.toString().replaceAll(" ", "%20");
    }
    
    /**
     * Just for not repeat code. Gets the correct separator according to the server URL
     * @param h
     * @return
     */
    private static String getSymbol(String h) {
        String symbol;
        if (h.indexOf("?")==-1) 
            symbol = "?";
        else if (h.indexOf("?")!=h.length()-1)
            symbol = "&";
        else
            symbol = "";
        return symbol;
    }

    /**
     * Gets the part of the OGC request that share GetMap and GetFeatureInfo
     * @return String request
     */
    public String getPartialQuery(WMSStatus status)
    {    	
        StringBuffer req = new StringBuffer();
        req.append("LAYERS=" + Utilities.Vector2CS(status.getLayerNames()))
           .append("&SRS=" + status.getSrs())
           .append("&BBOX=" + status.getExtent().getMinX()+ "," )
           .append(status.getExtent().getMinY()+ ",")
           .append(status.getExtent().getMaxX()+ ",")
           .append(status.getExtent().getMaxY())
           .append("&WIDTH=" + status.getWidth())
           .append("&HEIGHT=" + status.getHeight())
           .append("&FORMAT=" + status.getFormat())
           .append("&STYLES=");
        Vector v = status.getStyles();
        if (v!=null && v.size()>0)
        	req.append(Utilities.Vector2CS(v)); 
        v = status.getDimensions();
        if (v!=null && v.size()>0)
            req.append("&" + Utilities.Vector2URLParamString(v));
        if (status.getTransparency()) {
            req.append("&TRANSPARENT=TRUE");
        }
        return req.toString();
    }

    
    
    public void close() {        
        // your code here
    } 
    
    /**
     * Inner class that represents the description of the WMS metadata.
     * The first part of the capabilities will return the service information
     * from the WMS, this class will hold this information. 
     * 
     */
    public class ServiceInformation {

        public String online_resource = null;
        /*public String map_online_resource = null;
        public String feature_online_resource = null;*/
        public String version;
        public String name;
        public String scope;
        public String title;
        public String abstr;
        public String keywords;
        public String fees;
        public String operationsInfo;
        public String personname;
        public String organization;
        public String function;
        public String addresstype;
        public String address;
        public String place;
        public String province;
        public String postcode;
        public String country;
        public String phone;
        public String fax;
        public String email;
        public Vector formats;
        public HashMap operations; // operations that WMS supports
        
        public ServiceInformation()
        {  	
            version = new String();
            name = new String();
            scope = new String();
            title = new String();
            abstr = new String();
            keywords = new String();
            fees = new String();
            operationsInfo = new String();
            personname = new String();
            organization = new String();
            function = new String();
            addresstype = new String();
            address = new String();
            place = new String();
            province = new String();
            postcode = new String();
            country = new String();
            phone = new String();
            fax = new String();
            email = new String();
            formats = new Vector();       	
            operations = new HashMap();            
        }
        public boolean isQueryable()
        {
        	if (operations.keySet().contains( CapabilitiesTags.GETFEATUREINFO ))
        		return true;
        	else
        		return false;
        }
        public boolean hasLegendGraphic()
        {
        	if (operations.keySet().contains( CapabilitiesTags.GETLEGENDGRAPHIC))
        		return true;
        	else
        		return false;
        }        
        public void clear() {
        	version = new String();
            name = new String();
            scope = new String();
            title = new String();
            abstr = new String();
            keywords = new String();
            fees = new String();
            operationsInfo = new String();
            personname = new String();
            organization = new String();
            function = new String();
            addresstype = new String();
            address = new String();
            place = new String();
            province = new String();
            postcode = new String();
            country = new String();
            phone = new String();
            fax = new String();
            email = new String();
            formats = new Vector();       	
            operations = new HashMap();            
        }
     }   
 }
